from sysactions import (
    action,
    StopAction,
)

from application.interactor import TableInteractor


@action
def show_related_tabs(
        pos_id,    # type: str
        table_id   # type: str
):
    # type: (...) -> None

    from msgbusboundary import MsgBusTableService as tableService

    table_interactor = TableInteractor(pos_id, table_id, tableService())
    index = table_interactor.select_table_or_related_tab_by_dialog(pos_id, table_id)
    if index is None:
        raise StopAction()



from abc import ABCMeta, abstractmethod
from logging import Logger

from production.model import ProductionOrder
from typing import Dict, List, Optional, Union, Tuple


class ProductionBox(object):
    __metaclass__ = ABCMeta

    def __init__(self, name, sons, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], Logger) -> None
        self.name = name
        if sons is None:
            self.sons = []
        elif isinstance(sons, List):
            self.sons = sons
        else:
            self.sons = [sons]
        self.logger = logger
        self.refreshing = False
        self.tree_name = ''

        self.orders = {}  # type: Dict[Tuple[str, str], ProductionOrder]

    def set_tree_name(self, tree_name):
        # type: (str) -> None

        self.tree_name = tree_name

    @abstractmethod
    def order_modified(self, order):
        # type: (ProductionOrder) -> None

        raise NotImplementedError()

    def send_to_children(self, order):
        # type: (ProductionOrder) -> None

        if not hasattr(order, 'box_count'):
            order.box_count = 1

        order.box_count += 1

        if self.sons is not None:
            for son in self.sons:
                son.order_modified(order)

    def handle_production_command(self, command):
        if self.sons is not None:
            handled = False
            for son in self.sons:
                if son.handle_production_command(command):
                    handled = True
            return handled
        else:
            return False

    def start_refresh(self):
        self.refreshing = True
        if self.sons is not None:
            for son in self.sons:
                son.start_refresh()

    def end_refresh(self):
        if self.sons is not None:
            for son in self.sons:
                son.end_refresh()
        self.refreshing = False

    def get_order(self, order_id):
        # type: (str) -> Optional[ProductionOrder]

        if (self.name, order_id) in self.orders:
            return self.orders[(self.name, order_id)]

        return None

    def save_order(
        self,
        order,       # type: ProductionOrder
    ):
        # type: (...) -> None

        self.orders[(self.name, order.order_id)] = order

    def delete_order(self, order_id):
        # type: (str) -> None

        if (self.name, order_id) in self.orders:
            del self.orders[(self.name, order_id)]

    def debug(self, msg, *args):
        if self.logger is not None:
            try:
                if not args:
                    self.logger.debug(msg.decode("utf-8"))
                else:
                    self.logger.debug(msg.format(*args).decode("utf-8"))
            except Exception:
                msg = "Error formatting message"
                self.logger.error(msg)





#
# THIS IS WORK IN PROGRESS
#
# The Python Imaging Library.
# $Id$
#
# FlashPix support for PIL
#
# History:
# 97-01-25 fl   Created (reads uncompressed RGB images only)
#
# Copyright (c) Secret Labs AB 1997.
# Copyright (c) Fredrik Lundh 1997.
#
# See the README file for information on usage and redistribution.
#

from __future__ import print_function

import olefile

from . import Image, ImageFile
from ._binary import i8, i32le as i32

# __version__ is deprecated and will be removed in a future version. Use
# PIL.__version__ instead.
__version__ = "0.1"

# we map from colour field tuples to (mode, rawmode) descriptors
MODES = {
    # opacity
    (0x00007FFE): ("A", "L"),
    # monochrome
    (0x00010000,): ("L", "L"),
    (0x00018000, 0x00017FFE): ("RGBA", "LA"),
    # photo YCC
    (0x00020000, 0x00020001, 0x00020002): ("RGB", "YCC;P"),
    (0x00028000, 0x00028001, 0x00028002, 0x00027FFE): ("RGBA", "YCCA;P"),
    # standard RGB (NIFRGB)
    (0x00030000, 0x00030001, 0x00030002): ("RGB", "RGB"),
    (0x00038000, 0x00038001, 0x00038002, 0x00037FFE): ("RGBA", "RGBA"),
}


#
# --------------------------------------------------------------------


def _accept(prefix):
    return prefix[:8] == olefile.MAGIC


##
# Image plugin for the FlashPix images.


class FpxImageFile(ImageFile.ImageFile):

    format = "FPX"
    format_description = "FlashPix"

    def _open(self):
        #
        # read the OLE directory and see if this is a likely
        # to be a FlashPix file

        try:
            self.ole = olefile.OleFileIO(self.fp)
        except IOError:
            raise SyntaxError("not an FPX file; invalid OLE file")

        if self.ole.root.clsid != "56616700-C154-11CE-8553-00AA00A1F95B":
            raise SyntaxError("not an FPX file; bad root CLSID")

        self._open_index(1)

    def _open_index(self, index=1):
        #
        # get the Image Contents Property Set

        prop = self.ole.getproperties(
            ["Data Object Store %06d" % index, "\005Image Contents"]
        )

        # size (highest resolution)

        self._size = prop[0x1000002], prop[0x1000003]

        size = max(self.size)
        i = 1
        while size > 64:
            size = size / 2
            i += 1
        self.maxid = i - 1

        # mode.  instead of using a single field for this, flashpix
        # requires you to specify the mode for each channel in each
        # resolution subimage, and leaves it to the decoder to make
        # sure that they all match.  for now, we'll cheat and assume
        # that this is always the case.

        id = self.maxid << 16

        s = prop[0x2000002 | id]

        colors = []
        for i in range(i32(s, 4)):
            # note: for now, we ignore the "uncalibrated" flag
            colors.append(i32(s, 8 + i * 4) & 0x7FFFFFFF)

        self.mode, self.rawmode = MODES[tuple(colors)]

        # load JPEG tables, if any
        self.jpeg = {}
        for i in range(256):
            id = 0x3000001 | (i << 16)
            if id in prop:
                self.jpeg[i] = prop[id]

        self._open_subimage(1, self.maxid)

    def _open_subimage(self, index=1, subimage=0):
        #
        # setup tile descriptors for a given subimage

        stream = [
            "Data Object Store %06d" % index,
            "Resolution %04d" % subimage,
            "Subimage 0000 Header",
        ]

        fp = self.ole.openstream(stream)

        # skip prefix
        fp.read(28)

        # header stream
        s = fp.read(36)

        size = i32(s, 4), i32(s, 8)
        # tilecount = i32(s, 12)
        tilesize = i32(s, 16), i32(s, 20)
        # channels = i32(s, 24)
        offset = i32(s, 28)
        length = i32(s, 32)

        if size != self.size:
            raise IOError("subimage mismatch")

        # get tile descriptors
        fp.seek(28 + offset)
        s = fp.read(i32(s, 12) * length)

        x = y = 0
        xsize, ysize = size
        xtile, ytile = tilesize
        self.tile = []

        for i in range(0, len(s), length):

            compression = i32(s, i + 8)

            if compression == 0:
                self.tile.append(
                    (
                        "raw",
                        (x, y, x + xtile, y + ytile),
                        i32(s, i) + 28,
                        (self.rawmode),
                    )
                )

            elif compression == 1:

                # FIXME: the fill decoder is not implemented
                self.tile.append(
                    (
                        "fill",
                        (x, y, x + xtile, y + ytile),
                        i32(s, i) + 28,
                        (self.rawmode, s[12:16]),
                    )
                )

            elif compression == 2:

                internal_color_conversion = i8(s[14])
                jpeg_tables = i8(s[15])
                rawmode = self.rawmode

                if internal_color_conversion:
                    # The image is stored as usual (usually YCbCr).
                    if rawmode == "RGBA":
                        # For "RGBA", data is stored as YCbCrA based on
                        # negative RGB. The following trick works around
                        # this problem :
                        jpegmode, rawmode = "YCbCrK", "CMYK"
                    else:
                        jpegmode = None  # let the decoder decide

                else:
                    # The image is stored as defined by rawmode
                    jpegmode = rawmode

                self.tile.append(
                    (
                        "jpeg",
                        (x, y, x + xtile, y + ytile),
                        i32(s, i) + 28,
                        (rawmode, jpegmode),
                    )
                )

                # FIXME: jpeg tables are tile dependent; the prefix
                # data must be placed in the tile descriptor itself!

                if jpeg_tables:
                    self.tile_prefix = self.jpeg[jpeg_tables]

            else:
                raise IOError("unknown/invalid compression")

            x = x + xtile
            if x >= xsize:
                x, y = 0, y + ytile
                if y >= ysize:
                    break  # isn't really required

        self.stream = stream
        self.fp = None

    def load(self):

        if not self.fp:
            self.fp = self.ole.openstream(self.stream[:2] + ["Subimage 0000 Data"])

        return ImageFile.ImageFile.load(self)


#
# --------------------------------------------------------------------


Image.register_open(FpxImageFile.format, FpxImageFile, _accept)

Image.register_extension(FpxImageFile.format, ".fpx")



from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository

from persistence import (
    Connection,
    Cursor,
)

from model.services import NavigationServiceModel


class NavigationService(NavigationServiceModel, BaseRepository):

    def __init__(
        self,
        message_bus     # type: MbContextMessageBus
    ):
        # type: (...) -> None

        super(NavigationService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus

    def get_navigation_descriptions(
        self,
    ):
        # type: (...) -> str

        def inner_func(
            conn    # type: Connection
        ):
            # type: (...) -> dict

            query = """
                SELECT DescrId,
                    NavId,
                    NavigationDescription
                FROM NavigationDescriptions
            """
            cursor = conn.select(sql=query)
            navigations = self._parse_navigation_descriptions_data(cursor=cursor)

            return navigations

        descriptions_json = self.execute_with_connection(
            method=inner_func,
            service_name="Persistence",
        )

        return descriptions_json

    @staticmethod
    def _parse_navigation_descriptions_data(
        cursor,  # type: Cursor
    ):
        # type: (...) -> dict

        navigations = {}
        for row in cursor:
            descr_id = row.get_entry(0)
            nav_id = row.get_entry(1)
            description = row.get_entry(2)

            if nav_id not in navigations:
                navigations[nav_id] = {
                    "pt-br": "",
                    "en-us": "",
                    "es-es": "",
                }
            nav_descriptions = navigations.get(nav_id)

            language = None
            if descr_id == "4":
                language = "pt-br"
            elif descr_id == "14":
                language = "en-us"
            elif descr_id == "24":
                language = "es-es"

            if language:
                nav_descriptions[language] = description

        return navigations



"""
Utility for locating the module (or package's __init__.py)
associated with a given console_script name
and verifying it contains the PYTHON_ARGCOMPLETE_OK marker.

Such scripts are automatically generated and cannot contain
the marker themselves, so we defer to the containing module or package.

For more information on setuptools console_scripts, see
https://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation

Intended to be invoked by argcomplete's global completion function.
"""
import os
import sys

try:
    from importlib.metadata import entry_points as importlib_entry_points
except ImportError:
    from importlib_metadata import entry_points as importlib_entry_points

from ._check_module import ArgcompleteMarkerNotFound, find


def main():
    # Argument is the full path to the console script.
    script_path = sys.argv[1]

    # Find the module and function names that correspond to this
    # assuming it is actually a console script.
    name = os.path.basename(script_path)
    entry_points = [ep for ep in importlib_entry_points()["console_scripts"] if ep.name == name]
    if not entry_points:
        raise ArgcompleteMarkerNotFound('no entry point found matching script')
    entry_point = entry_points[0]
    module_name, function_name = entry_point.value.split(":", 1)

    # Check this looks like the script we really expected.
    with open(script_path) as f:
        script = f.read()
    if 'from {} import {}'.format(module_name, function_name) not in script:
        raise ArgcompleteMarkerNotFound('does not appear to be a console script')
    if 'sys.exit({}())'.format(function_name) not in script:
        raise ArgcompleteMarkerNotFound('does not appear to be a console script')

    # Look for the argcomplete marker in the script it imports.
    with open(find(module_name, return_package=True)) as f:
        head = f.read(1024)
    if 'PYTHON_ARGCOMPLETE_OK' not in head:
        raise ArgcompleteMarkerNotFound('marker not found')


if __name__ == '__main__':
    try:
        main()
    except ArgcompleteMarkerNotFound as e:
        sys.exit(e)



from production.box._OrderChangerProductionBox import OrderChangerProductionBox

from ._AllowedForbiddenFilter import AllowedForbiddenFilter


class TagFilterBox(OrderChangerProductionBox):
    def __init__(
        self,
        name,
        sons,
        allowed_tags,
        forbidden_tags,
        allow_combos,
        inherit_comments,
        logger=None
    ):
        super(TagFilterBox, self).__init__(name, sons, logger)
        self.filter = AllowedForbiddenFilter(
            allowed_values=allowed_tags,
            forbidden_values=forbidden_tags,
            keep_son_parts=allow_combos,
            inherit_comments=inherit_comments,
        )

    def change_order(self, order):
        order = self.filter.filter_items(order, lambda x: x.get_tags())
        return order



from iso8601 import iso8601
from typing import Dict


class Conditions(object):
    def __init__(self, obj):
        # type: (Dict) -> None

        self.user_min_level = obj.get("userMinLevel")
        self.unique_for_sale = bool(obj.get("uniqueForSale"))
        self.ignore_burned_items = bool(obj.get("ignoreBurnedItems"))
        self.single_use = bool(obj.get("singleUse")) or self.ignore_burned_items
        self.burn_items = bool(obj.get("burnItems", True))
        self.pod_type = obj.get("podType", None)
        self.min_sale_amount = obj.get("minSaleAmount", None)
        self.max_sale_amount = obj.get("maxSaleAmount", None)
        self.valid_start_period = obj.get("validStartPeriod", None)
        self.valid_end_period = obj.get("validEndPeriod", None)
        self.valid_hours = obj.get("validHours", None)
        self.valid_days = obj.get("validDays", None)
        self.required_items_on_sale = obj.get("requiredItemsOnSale", [])
        self.service_type = obj.get("serviceType", None)
        self.sale_type = obj.get("saleType", None)
        self.sector = obj.get("sector", None)
        self.min_seats = obj.get("minSeats")
        self.discount_max_value = obj.get("discountMaxValue")
        self.service_period = obj.get("servicePeriods", None)
        self.brands = obj.get("brands", None)
        self.required_groups_on_sale = obj.get("requiredGroupsOnSale", [])



from _AccountRepository import AccountRepository
from _OrderRepository import OrderRepository
from _FiscalRepository import FiscalRepository
from _TenderRepository import TenderRepository

__all__ = ['AccountRepository', 'OrderRepository', 'FiscalRepository', 'TenderRepository']



# -*- coding: utf-8 -*-
import base64

from requests import exceptions
from nfcecanceler import NfceCancelerResponseParser, NfceCancelerRequestBuilder, NfceCancelerResponse
from nfcebuilder.nfceutil import NfceRequest
import logging
import os

from typing import Dict

logger = logging.getLogger("FiscalWrapper")


class NfceCanceler(object):
    def __init__(self, nfce_canceler_request_builder, nfce_canceler_response_parser, nfce_request, url_cancelamento, dir_enviados, versao_ws):
        # type: (NfceCancelerRequestBuilder, NfceCancelerResponseParser, NfceRequest, str, str, int) -> None
        self.nfce_canceler_request_builder = nfce_canceler_request_builder
        self.nfce_canceler_response_parser = nfce_canceler_response_parser
        self.dir_enviados = dir_enviados
        self.nfce_request = nfce_request
        self.url_cancelamento = url_cancelamento
        self.versao_ws = versao_ws
        if versao_ws in (1, 3):
            self.soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/RecepcaoEvento/nfeRecepcaoEvento"
        else:
            self.soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeRecepcaoEvento4/nfeRecepcaoEvento"

    def cancel_order(
        self,
        order_response,     # type: str
        justification,      # type: str
        order_id,           # type: str
    ):
        # type: (...) -> Dict

        cancel_request_xml = self.nfce_canceler_request_builder.build_request(order_response, justification, self.versao_ws)

        try:
            response = self.nfce_request.envia_nfce(cancel_request_xml, self.url_cancelamento, soap_action=self.soap_act)
        except exceptions.RequestException:
            return {'status': False, 'message': "Problemas de Conexao", 'xml_response': None}
        except Exception as other_e:
            return {'status': False, 'message':  str(other_e), 'xml_response': None}
        else:
            nfce_canceler_response = self.nfce_canceler_response_parser.parse_response(response.text, self.versao_ws)
            if nfce_canceler_response.c_stat not in (NfceCancelerResponse.Canceled, NfceCancelerResponse.AlreadyCanceled):
                message = 'Invalid cStat {0}'.format(nfce_canceler_response.c_stat)
                return {'status': False, 'message':  message, 'xml_response': None}
            else:
                if nfce_canceler_response.c_stat == NfceCancelerResponse.Canceled:
                    idx_start = cancel_request_xml.find("<chNFe>") + 7
                    idx_final = cancel_request_xml.find("</chNFe>")
                    chave = cancel_request_xml[idx_start:idx_final]
                    serie_nota = chave[22:25]
                    numero_nota = chave[25:34].zfill(9)
                    idx_start = cancel_request_xml.find("<dhEvento>") + 10
                    idx_final = cancel_request_xml.find("</dhEvento>")
                    data_emissao = cancel_request_xml[idx_start:idx_final]
                    dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
                    name_template = "{}_{}_{}_nfe_cancelamento_NFe{}.xml"
                    file_name = name_template.format(serie_nota, numero_nota, order_id.zfill(9), chave)
                    file_path = os.path.join(self.dir_enviados, "Enviados", dir_arquivo, file_name)
                    with open(file_path, "w+") as nfe_cancel_file:
                        nfe_cancel_file.write(response.text)

                    xml_response = base64.b64encode(response.text)
                    return {'status': True, 'message': "Venda cancelada com sucesso", 'xml_response': xml_response}
                elif nfce_canceler_response.c_stat == NfceCancelerResponse.AlreadyCanceled:
                    return {'status': True, 'message': "Venda já cancelada anteriormente", 'xml_response': None}
                else:
                    return {'status': False, 'message': "Erro ao tentar cancelar venda", 'xml_response': None}



from application.model.configuration import Configurations
from application.model.customexception import build_message_exception
from application.repository.api import APIRepository
from typing import Optional


class GetLoyaltyIdInteractor(object):
    def __init__(self, configs, api_repository):
        # type: (Configurations, APIRepository) -> None

        self.logger = configs.logger
        self.configs = configs
        self.api_repository = api_repository

    def get_loyalty_id(self, loyalty_customer_id):
        # type: (str) -> Optional[str]

        try:
            self.logger.info("Requesting customer loyaltySystemId")
            loyalty_id = self.api_repository.retrieve_loyalty_customer_info(loyalty_customer_id)
            if not loyalty_id:
                self.logger.info("No customer loyalty data retrieved")
            else:
                self.logger.info("Retrieved customer loyalty data: {}".format(loyalty_id))

            return loyalty_id
        except Exception as ex:
            self.logger.error("Error at get_loyalty_id: {}".format(ex))
            build_message_exception(ex)
            raise



from _OrderRepository import OrderRepository
from _ProductRepository import ProductRepository
from _AccountRepository import AccountRepository

__all__ = ["OrderRepository", "ProductRepository", "AccountRepository"]



from __future__ import absolute_import

import io
import logging
import sys
import warnings
import zlib
from contextlib import contextmanager
from socket import error as SocketError
from socket import timeout as SocketTimeout

try:
    try:
        import brotlicffi as brotli
    except ImportError:
        import brotli
except ImportError:
    brotli = None

from . import util
from ._collections import HTTPHeaderDict
from .connection import BaseSSLError, HTTPException
from .exceptions import (
    BodyNotHttplibCompatible,
    DecodeError,
    HTTPError,
    IncompleteRead,
    InvalidChunkLength,
    InvalidHeader,
    ProtocolError,
    ReadTimeoutError,
    ResponseNotChunked,
    SSLError,
)
from .packages import six
from .util.response import is_fp_closed, is_response_to_head

log = logging.getLogger(__name__)


class DeflateDecoder(object):
    def __init__(self):
        self._first_try = True
        self._data = b""
        self._obj = zlib.decompressobj()

    def __getattr__(self, name):
        return getattr(self._obj, name)

    def decompress(self, data):
        if not data:
            return data

        if not self._first_try:
            return self._obj.decompress(data)

        self._data += data
        try:
            decompressed = self._obj.decompress(data)
            if decompressed:
                self._first_try = False
                self._data = None
            return decompressed
        except zlib.error:
            self._first_try = False
            self._obj = zlib.decompressobj(-zlib.MAX_WBITS)
            try:
                return self.decompress(self._data)
            finally:
                self._data = None


class GzipDecoderState(object):

    FIRST_MEMBER = 0
    OTHER_MEMBERS = 1
    SWALLOW_DATA = 2


class GzipDecoder(object):
    def __init__(self):
        self._obj = zlib.decompressobj(16 + zlib.MAX_WBITS)
        self._state = GzipDecoderState.FIRST_MEMBER

    def __getattr__(self, name):
        return getattr(self._obj, name)

    def decompress(self, data):
        ret = bytearray()
        if self._state == GzipDecoderState.SWALLOW_DATA or not data:
            return bytes(ret)
        while True:
            try:
                ret += self._obj.decompress(data)
            except zlib.error:
                previous_state = self._state
                # Ignore data after the first error
                self._state = GzipDecoderState.SWALLOW_DATA
                if previous_state == GzipDecoderState.OTHER_MEMBERS:
                    # Allow trailing garbage acceptable in other gzip clients
                    return bytes(ret)
                raise
            data = self._obj.unused_data
            if not data:
                return bytes(ret)
            self._state = GzipDecoderState.OTHER_MEMBERS
            self._obj = zlib.decompressobj(16 + zlib.MAX_WBITS)


if brotli is not None:

    class BrotliDecoder(object):
        # Supports both 'brotlipy' and 'Brotli' packages
        # since they share an import name. The top branches
        # are for 'brotlipy' and bottom branches for 'Brotli'
        def __init__(self):
            self._obj = brotli.Decompressor()
            if hasattr(self._obj, "decompress"):
                self.decompress = self._obj.decompress
            else:
                self.decompress = self._obj.process

        def flush(self):
            if hasattr(self._obj, "flush"):
                return self._obj.flush()
            return b""


class MultiDecoder(object):
    """
    From RFC7231:
        If one or more encodings have been applied to a representation, the
        sender that applied the encodings MUST generate a Content-Encoding
        header field that lists the content codings in the order in which
        they were applied.
    """

    def __init__(self, modes):
        self._decoders = [_get_decoder(m.strip()) for m in modes.split(",")]

    def flush(self):
        return self._decoders[0].flush()

    def decompress(self, data):
        for d in reversed(self._decoders):
            data = d.decompress(data)
        return data


def _get_decoder(mode):
    if "," in mode:
        return MultiDecoder(mode)

    if mode == "gzip":
        return GzipDecoder()

    if brotli is not None and mode == "br":
        return BrotliDecoder()

    return DeflateDecoder()


class HTTPResponse(io.IOBase):
    """
    HTTP Response container.

    Backwards-compatible with :class:`http.client.HTTPResponse` but the response ``body`` is
    loaded and decoded on-demand when the ``data`` property is accessed.  This
    class is also compatible with the Python standard library's :mod:`io`
    module, and can hence be treated as a readable object in the context of that
    framework.

    Extra parameters for behaviour not present in :class:`http.client.HTTPResponse`:

    :param preload_content:
        If True, the response's body will be preloaded during construction.

    :param decode_content:
        If True, will attempt to decode the body based on the
        'content-encoding' header.

    :param original_response:
        When this HTTPResponse wrapper is generated from an :class:`http.client.HTTPResponse`
        object, it's convenient to include the original for debug purposes. It's
        otherwise unused.

    :param retries:
        The retries contains the last :class:`~urllib3.util.retry.Retry` that
        was used during the request.

    :param enforce_content_length:
        Enforce content length checking. Body returned by server must match
        value of Content-Length header, if present. Otherwise, raise error.
    """

    CONTENT_DECODERS = ["gzip", "deflate"]
    if brotli is not None:
        CONTENT_DECODERS += ["br"]
    REDIRECT_STATUSES = [301, 302, 303, 307, 308]

    def __init__(
        self,
        body="",
        headers=None,
        status=0,
        version=0,
        reason=None,
        strict=0,
        preload_content=True,
        decode_content=True,
        original_response=None,
        pool=None,
        connection=None,
        msg=None,
        retries=None,
        enforce_content_length=False,
        request_method=None,
        request_url=None,
        auto_close=True,
    ):

        if isinstance(headers, HTTPHeaderDict):
            self.headers = headers
        else:
            self.headers = HTTPHeaderDict(headers)
        self.status = status
        self.version = version
        self.reason = reason
        self.strict = strict
        self.decode_content = decode_content
        self.retries = retries
        self.enforce_content_length = enforce_content_length
        self.auto_close = auto_close

        self._decoder = None
        self._body = None
        self._fp = None
        self._original_response = original_response
        self._fp_bytes_read = 0
        self.msg = msg
        self._request_url = request_url

        if body and isinstance(body, (six.string_types, bytes)):
            self._body = body

        self._pool = pool
        self._connection = connection

        if hasattr(body, "read"):
            self._fp = body

        # Are we using the chunked-style of transfer encoding?
        self.chunked = False
        self.chunk_left = None
        tr_enc = self.headers.get("transfer-encoding", "").lower()
        # Don't incur the penalty of creating a list and then discarding it
        encodings = (enc.strip() for enc in tr_enc.split(","))
        if "chunked" in encodings:
            self.chunked = True

        # Determine length of response
        self.length_remaining = self._init_length(request_method)

        # If requested, preload the body.
        if preload_content and not self._body:
            self._body = self.read(decode_content=decode_content)

    def get_redirect_location(self):
        """
        Should we redirect and where to?

        :returns: Truthy redirect location string if we got a redirect status
            code and valid location. ``None`` if redirect status and no
            location. ``False`` if not a redirect status code.
        """
        if self.status in self.REDIRECT_STATUSES:
            return self.headers.get("location")

        return False

    def release_conn(self):
        if not self._pool or not self._connection:
            return

        self._pool._put_conn(self._connection)
        self._connection = None

    def drain_conn(self):
        """
        Read and discard any remaining HTTP response data in the response connection.

        Unread data in the HTTPResponse connection blocks the connection from being released back to the pool.
        """
        try:
            self.read()
        except (HTTPError, SocketError, BaseSSLError, HTTPException):
            pass

    @property
    def data(self):
        # For backwards-compat with earlier urllib3 0.4 and earlier.
        if self._body:
            return self._body

        if self._fp:
            return self.read(cache_content=True)

    @property
    def connection(self):
        return self._connection

    def isclosed(self):
        return is_fp_closed(self._fp)

    def tell(self):
        """
        Obtain the number of bytes pulled over the wire so far. May differ from
        the amount of content returned by :meth:``urllib3.response.HTTPResponse.read``
        if bytes are encoded on the wire (e.g, compressed).
        """
        return self._fp_bytes_read

    def _init_length(self, request_method):
        """
        Set initial length value for Response content if available.
        """
        length = self.headers.get("content-length")

        if length is not None:
            if self.chunked:
                # This Response will fail with an IncompleteRead if it can't be
                # received as chunked. This method falls back to attempt reading
                # the response before raising an exception.
                log.warning(
                    "Received response with both Content-Length and "
                    "Transfer-Encoding set. This is expressly forbidden "
                    "by RFC 7230 sec 3.3.2. Ignoring Content-Length and "
                    "attempting to process response as Transfer-Encoding: "
                    "chunked."
                )
                return None

            try:
                # RFC 7230 section 3.3.2 specifies multiple content lengths can
                # be sent in a single Content-Length header
                # (e.g. Content-Length: 42, 42). This line ensures the values
                # are all valid ints and that as long as the `set` length is 1,
                # all values are the same. Otherwise, the header is invalid.
                lengths = set([int(val) for val in length.split(",")])
                if len(lengths) > 1:
                    raise InvalidHeader(
                        "Content-Length contained multiple "
                        "unmatching values (%s)" % length
                    )
                length = lengths.pop()
            except ValueError:
                length = None
            else:
                if length < 0:
                    length = None

        # Convert status to int for comparison
        # In some cases, httplib returns a status of "_UNKNOWN"
        try:
            status = int(self.status)
        except ValueError:
            status = 0

        # Check for responses that shouldn't include a body
        if status in (204, 304) or 100 <= status < 200 or request_method == "HEAD":
            length = 0

        return length

    def _init_decoder(self):
        """
        Set-up the _decoder attribute if necessary.
        """
        # Note: content-encoding value should be case-insensitive, per RFC 7230
        # Section 3.2
        content_encoding = self.headers.get("content-encoding", "").lower()
        if self._decoder is None:
            if content_encoding in self.CONTENT_DECODERS:
                self._decoder = _get_decoder(content_encoding)
            elif "," in content_encoding:
                encodings = [
                    e.strip()
                    for e in content_encoding.split(",")
                    if e.strip() in self.CONTENT_DECODERS
                ]
                if len(encodings):
                    self._decoder = _get_decoder(content_encoding)

    DECODER_ERROR_CLASSES = (IOError, zlib.error)
    if brotli is not None:
        DECODER_ERROR_CLASSES += (brotli.error,)

    def _decode(self, data, decode_content, flush_decoder):
        """
        Decode the data passed in and potentially flush the decoder.
        """
        if not decode_content:
            return data

        try:
            if self._decoder:
                data = self._decoder.decompress(data)
        except self.DECODER_ERROR_CLASSES as e:
            content_encoding = self.headers.get("content-encoding", "").lower()
            raise DecodeError(
                "Received response with content-encoding: %s, but "
                "failed to decode it." % content_encoding,
                e,
            )
        if flush_decoder:
            data += self._flush_decoder()

        return data

    def _flush_decoder(self):
        """
        Flushes the decoder. Should only be called if the decoder is actually
        being used.
        """
        if self._decoder:
            buf = self._decoder.decompress(b"")
            return buf + self._decoder.flush()

        return b""

    @contextmanager
    def _error_catcher(self):
        """
        Catch low-level python exceptions, instead re-raising urllib3
        variants, so that low-level exceptions are not leaked in the
        high-level api.

        On exit, release the connection back to the pool.
        """
        clean_exit = False

        try:
            try:
                yield

            except SocketTimeout:
                # FIXME: Ideally we'd like to include the url in the ReadTimeoutError but
                # there is yet no clean way to get at it from this context.
                raise ReadTimeoutError(self._pool, None, "Read timed out.")

            except BaseSSLError as e:
                # FIXME: Is there a better way to differentiate between SSLErrors?
                if "read operation timed out" not in str(e):
                    # SSL errors related to framing/MAC get wrapped and reraised here
                    raise SSLError(e)

                raise ReadTimeoutError(self._pool, None, "Read timed out.")

            except (HTTPException, SocketError) as e:
                # This includes IncompleteRead.
                raise ProtocolError("Connection broken: %r" % e, e)

            # If no exception is thrown, we should avoid cleaning up
            # unnecessarily.
            clean_exit = True
        finally:
            # If we didn't terminate cleanly, we need to throw away our
            # connection.
            if not clean_exit:
                # The response may not be closed but we're not going to use it
                # anymore so close it now to ensure that the connection is
                # released back to the pool.
                if self._original_response:
                    self._original_response.close()

                # Closing the response may not actually be sufficient to close
                # everything, so if we have a hold of the connection close that
                # too.
                if self._connection:
                    self._connection.close()

            # If we hold the original response but it's closed now, we should
            # return the connection back to the pool.
            if self._original_response and self._original_response.isclosed():
                self.release_conn()

    def _fp_read(self, amt):
        """
        Read a response with the thought that reading the number of bytes
        larger than can fit in a 32-bit int at a time via SSL in some
        known cases leads to an overflow error that has to be prevented
        if `amt` or `self.length_remaining` indicate that a problem may
        happen.

        The known cases:
          * 3.8 <= CPython < 3.9.7 because of a bug
            https://github.com/urllib3/urllib3/issues/2513#issuecomment-1152559900.
          * urllib3 injected with pyOpenSSL-backed SSL-support.
          * CPython < 3.10 only when `amt` does not fit 32-bit int.
        """
        assert self._fp
        c_int_max = 2 ** 31 - 1
        if (
            (
                (amt and amt > c_int_max)
                or (self.length_remaining and self.length_remaining > c_int_max)
            )
            and not util.IS_SECURETRANSPORT
            and (util.IS_PYOPENSSL or sys.version_info < (3, 10))
        ):
            buffer = io.BytesIO()
            # Besides `max_chunk_amt` being a maximum chunk size, it
            # affects memory overhead of reading a response by this
            # method in CPython.
            # `c_int_max` equal to 2 GiB - 1 byte is the actual maximum
            # chunk size that does not lead to an overflow error, but
            # 256 MiB is a compromise.
            max_chunk_amt = 2 ** 28
            while amt is None or amt != 0:
                if amt is not None:
                    chunk_amt = min(amt, max_chunk_amt)
                    amt -= chunk_amt
                else:
                    chunk_amt = max_chunk_amt
                data = self._fp.read(chunk_amt)
                if not data:
                    break
                buffer.write(data)
                del data  # to reduce peak memory usage by `max_chunk_amt`.
            return buffer.getvalue()
        else:
            # StringIO doesn't like amt=None
            return self._fp.read(amt) if amt is not None else self._fp.read()

    def read(self, amt=None, decode_content=None, cache_content=False):
        """
        Similar to :meth:`http.client.HTTPResponse.read`, but with two additional
        parameters: ``decode_content`` and ``cache_content``.

        :param amt:
            How much of the content to read. If specified, caching is skipped
            because it doesn't make sense to cache partial content as the full
            response.

        :param decode_content:
            If True, will attempt to decode the body based on the
            'content-encoding' header.

        :param cache_content:
            If True, will save the returned data such that the same result is
            returned despite of the state of the underlying file object. This
            is useful if you want the ``.data`` property to continue working
            after having ``.read()`` the file object. (Overridden if ``amt`` is
            set.)
        """
        self._init_decoder()
        if decode_content is None:
            decode_content = self.decode_content

        if self._fp is None:
            return

        flush_decoder = False
        fp_closed = getattr(self._fp, "closed", False)

        with self._error_catcher():
            data = self._fp_read(amt) if not fp_closed else b""
            if amt is None:
                flush_decoder = True
            else:
                cache_content = False
                if (
                    amt != 0 and not data
                ):  # Platform-specific: Buggy versions of Python.
                    # Close the connection when no data is returned
                    #
                    # This is redundant to what httplib/http.client _should_
                    # already do.  However, versions of python released before
                    # December 15, 2012 (http://bugs.python.org/issue16298) do
                    # not properly close the connection in all cases. There is
                    # no harm in redundantly calling close.
                    self._fp.close()
                    flush_decoder = True
                    if self.enforce_content_length and self.length_remaining not in (
                        0,
                        None,
                    ):
                        # This is an edge case that httplib failed to cover due
                        # to concerns of backward compatibility. We're
                        # addressing it here to make sure IncompleteRead is
                        # raised during streaming, so all calls with incorrect
                        # Content-Length are caught.
                        raise IncompleteRead(self._fp_bytes_read, self.length_remaining)

        if data:
            self._fp_bytes_read += len(data)
            if self.length_remaining is not None:
                self.length_remaining -= len(data)

            data = self._decode(data, decode_content, flush_decoder)

            if cache_content:
                self._body = data

        return data

    def stream(self, amt=2 ** 16, decode_content=None):
        """
        A generator wrapper for the read() method. A call will block until
        ``amt`` bytes have been read from the connection or until the
        connection is closed.

        :param amt:
            How much of the content to read. The generator will return up to
            much data per iteration, but may return less. This is particularly
            likely when using compressed data. However, the empty string will
            never be returned.

        :param decode_content:
            If True, will attempt to decode the body based on the
            'content-encoding' header.
        """
        if self.chunked and self.supports_chunked_reads():
            for line in self.read_chunked(amt, decode_content=decode_content):
                yield line
        else:
            while not is_fp_closed(self._fp):
                data = self.read(amt=amt, decode_content=decode_content)

                if data:
                    yield data

    @classmethod
    def from_httplib(ResponseCls, r, **response_kw):
        """
        Given an :class:`http.client.HTTPResponse` instance ``r``, return a
        corresponding :class:`urllib3.response.HTTPResponse` object.

        Remaining parameters are passed to the HTTPResponse constructor, along
        with ``original_response=r``.
        """
        headers = r.msg

        if not isinstance(headers, HTTPHeaderDict):
            if six.PY2:
                # Python 2.7
                headers = HTTPHeaderDict.from_httplib(headers)
            else:
                headers = HTTPHeaderDict(headers.items())

        # HTTPResponse objects in Python 3 don't have a .strict attribute
        strict = getattr(r, "strict", 0)
        resp = ResponseCls(
            body=r,
            headers=headers,
            status=r.status,
            version=r.version,
            reason=r.reason,
            strict=strict,
            original_response=r,
            **response_kw
        )
        return resp

    # Backwards-compatibility methods for http.client.HTTPResponse
    def getheaders(self):
        warnings.warn(
            "HTTPResponse.getheaders() is deprecated and will be removed "
            "in urllib3 v2.1.0. Instead access HTTPResponse.headers directly.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.headers

    def getheader(self, name, default=None):
        warnings.warn(
            "HTTPResponse.getheader() is deprecated and will be removed "
            "in urllib3 v2.1.0. Instead use HTTPResponse.headers.get(name, default).",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.headers.get(name, default)

    # Backwards compatibility for http.cookiejar
    def info(self):
        return self.headers

    # Overrides from io.IOBase
    def close(self):
        if not self.closed:
            self._fp.close()

        if self._connection:
            self._connection.close()

        if not self.auto_close:
            io.IOBase.close(self)

    @property
    def closed(self):
        if not self.auto_close:
            return io.IOBase.closed.__get__(self)
        elif self._fp is None:
            return True
        elif hasattr(self._fp, "isclosed"):
            return self._fp.isclosed()
        elif hasattr(self._fp, "closed"):
            return self._fp.closed
        else:
            return True

    def fileno(self):
        if self._fp is None:
            raise IOError("HTTPResponse has no file to get a fileno from")
        elif hasattr(self._fp, "fileno"):
            return self._fp.fileno()
        else:
            raise IOError(
                "The file-like object this HTTPResponse is wrapped "
                "around has no file descriptor"
            )

    def flush(self):
        if (
            self._fp is not None
            and hasattr(self._fp, "flush")
            and not getattr(self._fp, "closed", False)
        ):
            return self._fp.flush()

    def readable(self):
        # This method is required for `io` module compatibility.
        return True

    def readinto(self, b):
        # This method is required for `io` module compatibility.
        temp = self.read(len(b))
        if len(temp) == 0:
            return 0
        else:
            b[: len(temp)] = temp
            return len(temp)

    def supports_chunked_reads(self):
        """
        Checks if the underlying file-like object looks like a
        :class:`http.client.HTTPResponse` object. We do this by testing for
        the fp attribute. If it is present we assume it returns raw chunks as
        processed by read_chunked().
        """
        return hasattr(self._fp, "fp")

    def _update_chunk_length(self):
        # First, we'll figure out length of a chunk and then
        # we'll try to read it from socket.
        if self.chunk_left is not None:
            return
        line = self._fp.fp.readline()
        line = line.split(b";", 1)[0]
        try:
            self.chunk_left = int(line, 16)
        except ValueError:
            # Invalid chunked protocol response, abort.
            self.close()
            raise InvalidChunkLength(self, line)

    def _handle_chunk(self, amt):
        returned_chunk = None
        if amt is None:
            chunk = self._fp._safe_read(self.chunk_left)
            returned_chunk = chunk
            self._fp._safe_read(2)  # Toss the CRLF at the end of the chunk.
            self.chunk_left = None
        elif amt < self.chunk_left:
            value = self._fp._safe_read(amt)
            self.chunk_left = self.chunk_left - amt
            returned_chunk = value
        elif amt == self.chunk_left:
            value = self._fp._safe_read(amt)
            self._fp._safe_read(2)  # Toss the CRLF at the end of the chunk.
            self.chunk_left = None
            returned_chunk = value
        else:  # amt > self.chunk_left
            returned_chunk = self._fp._safe_read(self.chunk_left)
            self._fp._safe_read(2)  # Toss the CRLF at the end of the chunk.
            self.chunk_left = None
        return returned_chunk

    def read_chunked(self, amt=None, decode_content=None):
        """
        Similar to :meth:`HTTPResponse.read`, but with an additional
        parameter: ``decode_content``.

        :param amt:
            How much of the content to read. If specified, caching is skipped
            because it doesn't make sense to cache partial content as the full
            response.

        :param decode_content:
            If True, will attempt to decode the body based on the
            'content-encoding' header.
        """
        self._init_decoder()
        # FIXME: Rewrite this method and make it a class with a better structured logic.
        if not self.chunked:
            raise ResponseNotChunked(
                "Response is not chunked. "
                "Header 'transfer-encoding: chunked' is missing."
            )
        if not self.supports_chunked_reads():
            raise BodyNotHttplibCompatible(
                "Body should be http.client.HTTPResponse like. "
                "It should have have an fp attribute which returns raw chunks."
            )

        with self._error_catcher():
            # Don't bother reading the body of a HEAD request.
            if self._original_response and is_response_to_head(self._original_response):
                self._original_response.close()
                return

            # If a response is already read and closed
            # then return immediately.
            if self._fp.fp is None:
                return

            while True:
                self._update_chunk_length()
                if self.chunk_left == 0:
                    break
                chunk = self._handle_chunk(amt)
                decoded = self._decode(
                    chunk, decode_content=decode_content, flush_decoder=False
                )
                if decoded:
                    yield decoded

            if decode_content:
                # On CPython and PyPy, we should never need to flush the
                # decoder. However, on Jython we *might* need to, so
                # lets defensively do it anyway.
                decoded = self._flush_decoder()
                if decoded:  # Platform-specific: Jython.
                    yield decoded

            # Chunk content ends with \r\n: discard it.
            while True:
                line = self._fp.fp.readline()
                if not line:
                    # Some sites may not end with '\r\n'.
                    break
                if line == b"\r\n":
                    break

            # We read everything; close the "file".
            if self._original_response:
                self._original_response.close()

    def geturl(self):
        """
        Returns the URL that was the source of this response.
        If the request that generated this response redirected, this method
        will return the final redirect location.
        """
        if self.retries is not None and len(self.retries.history):
            return self.retries.history[-1].redirect_location
        else:
            return self._request_url

    def __iter__(self):
        buffer = []
        for chunk in self.stream(decode_content=True):
            if b"\n" in chunk:
                chunk = chunk.split(b"\n")
                yield b"".join(buffer) + chunk[0] + b"\n"
                for x in chunk[1:-1]:
                    yield x + b"\n"
                if chunk[-1]:
                    buffer = [chunk[-1]]
                else:
                    buffer = []
            else:
                buffer.append(chunk)
        if buffer:
            yield b"".join(buffer)



import json
import os
import threading
from datetime import (
    datetime,
    timedelta,
)
from logging import Logger
from typing import (
    Dict,
    List,
)

from msgbus import MBEasyContext
from mwhelper import BaseRepository
from persistence import (
    Connection,
    Row,
)

from services.model import (
    Configs,
    ConfiguredAlert,
    Event,
)


lock_db = threading.Lock()


class AlertsRepository(BaseRepository):

    def __init__(
        self,
        mb_context,  # type: MBEasyContext
        configs,     # type: Configs
        logger,      # type: Logger
    ):
        # type: (...) -> None

        self.service_name = "AlertsPersistence"
        self.configs = configs
        self.logger = logger
        self.alerts_db_exported_xml = "alerts.db.xml"

        super(
            AlertsRepository,
            self,
        ).__init__(mb_context)

    def get_configured_alerts(
        self,
    ):
        # type: () -> List[ConfiguredAlert]

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> List[ConfiguredAlert]

            query = """
                SELECT Id,
                    AlertJson
                FROM alerts.ConfiguredAlerts
            """
            configured_alerts = []
            for row in conn.select(sql=query):
                obj = self._create_configured_alert(row)
                configured_alerts.append(obj)

            return configured_alerts

        result = self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name
        )

        return result

    def update_configured_alerts(
        self,
        configured_alerts,  # type: List[Dict]
    ):
        # type: (...) -> None

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> None

            try:
                conn.query("DELETE FROM alerts.ConfiguredAlerts;")
                self._update_configured_alerts(conn, configured_alerts)
            except Exception:  # noqa
                self.logger.exception("Error updating configured alerts")
                raise
            else:
                self.logger.info("Configured alerts updated successfully")

        self.execute_with_transaction(
            method=inner_func,
            service_name=self.service_name
        )

    def get_events(
        self,
        range_days=None,  # type: int
    ):
        # type: (...) -> List[Event]

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> List[Event]

            query = """
                SELECT 
                    e.Id,
                    e.AlertId,
                    e.NotificationJson,
                    e.Timestamp,
                    e.SentStatus,
                    s.TypeDescr
                FROM alerts.Events e
                JOIN alerts.StatusType s ON s.TypeId = e.SentStatus
            """

            if range_days is not None:
                query += """
                    WHERE e.Timestamp <= '{}';
                """
                query = query.format(str(datetime.utcnow() - timedelta(days=range_days)))

            events = []
            for row in conn.select(sql=query):
                obj = self._create_event(row)
                events.append(obj)

            return events

        result = self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name,
        )

        return result

    def get_events_with_sent_status(
        self,
        sent_status,   # type: int
    ):
        # type: (...) -> List[Dict]

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> List[Dict]

            query = """
                SELECT Id, NotificationJson, Timestamp, SentStatus
                FROM alerts.Events
                WHERE SentStatus='{}';"""
            events = []
            for row in conn.select(sql=query.format(sent_status)):
                events.append({
                    'id': row.get_entry('Id'),
                    'notification_json': row.get_entry('NotificationJson'),
                    'time_stamp': row.get_entry('Timestamp'),
                    'sent_status': row.get_entry('SentStatus'),
                })

            return events

        result = self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name,
        )

        return result

    def set_sent_status(
        self,
        event_id,      # type: str
        sent_status,   # type: int
    ):
        # type: (...) -> None

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> None

            query = """
                UPDATE alerts.Events
                set SentStatus='{}' 
                WHERE Id='{}';"""

            conn.query(sql=query.format(
                sent_status,
                event_id,
            ))

        self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name,
        )

    def _update_notification_json(
        self,
        event_id,           # type: str
        notification_json,  # type: dict
    ):
        # type: (...) -> None

        notification_json.update({
            "id": event_id,
        })

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> None

            query = """
                UPDATE alerts.Events
                set NotificationJson='{}' 
                WHERE Id='{}';"""

            conn.query(sql=query.format(
                json.dumps(notification_json),
                event_id,
            ))

        self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name,
        )

    def insert_event(
        self,
        alert_id,           # type: str
        notification_json,  # type: dict
        timestamp,          # type: str
        sent_status=0,      # type: int
    ):
        # type: (...) -> None

        with lock_db:
            def inner_func(
                conn,  # type: Connection
            ):
                # type: (...) -> str

                query = """
                    INSERT INTO alerts.Events (AlertId, NotificationJson, Timestamp, SentStatus)
                    VALUES ('{}', '{}', '{}', '{}');
                """
                conn.query(
                    query.format(
                        alert_id,
                        json.dumps(notification_json),
                        timestamp,
                        sent_status
                    )
                )

                query = """
                    SELECT MAX(id)
                    FROM Events;
                """
                cursor = conn.select(sql=query)
                row_id = cursor.get_row(0).get_entry(0)

                return row_id

            event_id = self.execute_with_connection(
                method=inner_func,
                service_name=self.service_name,
            )

            self._update_notification_json(
                event_id=event_id,
                notification_json=notification_json,
            )

    def delete_outdated_events(
        self,
        range_days,  # type: int
    ):
        # type: (...) -> None

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> bool

            try:
                query = """
                    DELETE FROM alerts.Events
                    WHERE Timestamp <= '{}';
                """
                query = query.format(str(datetime.utcnow() - timedelta(days=range_days)))
                conn.query(sql=query)

            except Exception as _:
                return False

            return True

        result = self.execute_with_connection(
            method=inner_func,
            service_name=self.service_name,
        )

        if not result:
            raise Exception("Error deleting outdated events")

    def vacuum_database(
        self,
    ):
        # type: (...) -> None

        with lock_db:
            db_info = {}

            def inner_func(
                conn,  # type: Connection
            ):
                # type: (...) -> bool

                try:
                    db_info["db_file"] = self.get_db_file(
                        conn=conn,
                        db_name_to_find="alerts",
                    )
                    db_info["old_file_size"] = self.get_file_size(file_name=db_info.get("db_file"))
                    query = """
                        VACUUM alerts;
                    """
                    conn.query(sql=query)
                except Exception as _:
                    return False

                return True

            result = self.execute_with_connection(
                method=inner_func,
                service_name=self.service_name,
            )

            if not result:
                raise Exception("Error doing vacuum on alerts db")

            db_info["new_file_size"] = self.get_file_size(file_name=db_info.get("db_file"))
            self.logger.info("Alerts db file sizes. OLD: {} bytes; NEW: {} bytes".format(
                db_info.get("old_file_size"),
                db_info.get("new_file_size"),
            ))
            self.logger.info("Finished [vacuum_alerts_database]")

    @staticmethod
    def _update_configured_alerts(
        conn,               # type: Connection
        configured_alerts,  # type: List[Dict]
    ):
        # type: (...) -> None

        for index, alert in enumerate(configured_alerts, 1):
            query = """
                INSERT INTO alerts.ConfiguredAlerts (Id, AlertJson)
                VALUES ('{}', '{}')
            """
            conn.query(
                query.format(
                    index,
                    json.dumps(alert)
                )
            )

    @staticmethod
    def get_db_file(
        conn,
        db_name_to_find,
    ):

        db_file = None
        pragma_cursor = conn.select("PRAGMA database_list;")
        for pragma_row in pragma_cursor:
            db_name, db_file = map(pragma_row.get_entry, (1, 2,))
            if db_name == db_name_to_find:
                break
        return db_file

    @staticmethod
    def get_file_size(
        file_name,  # type: str
    ):
        # type: (...) -> int

        return os.path.getsize(file_name)

    @staticmethod
    def _create_configured_alert(
        row,  # type: Row
    ):
        # type: (...) -> ConfiguredAlert

        obj = ConfiguredAlert(
            alert_id=row.get_entry("Id"),
            alert_json=json.loads(row.get_entry("AlertJson")),
        )
        return obj

    @staticmethod
    def _create_event(
        row,  # type: Row
    ):
        # type: (...) -> Event

        obj = Event(
            event_id=row.get_entry("Id"),
            alert_id=row.get_entry("AlertId"),
            notification_json=row.get_entry("NotificationJson"),
            timestamp=row.get_entry("Timestamp"),
            sent_status=row.get_entry("SentStatus"),
            status_desc=row.get_entry("TypeDescr"),
        )
        return obj



# encoding: utf-8


class JustificationHeaderDto(object):
    def __init__(
        self,
        title,      # type: str
        subtitle,   # type: str
    ):
        # type: (...) -> None

        self.title = title
        self.subtitle = subtitle



from pos_model import OrderState
from typing import List  # noqa

from enum import Enum
from _OrderTender import OrderTender  # noqa


class Order(object):
    def __init__(self, state, total, tenders):
        # type: (OrderState, float, List[OrderTender]) -> None
        self.state = state
        self.total = total
        self.tenders = tenders

    class OrderSate(Enum):
        voided = 1
        paid = 2




from ._TipReportDto import TipReportDto
from ._TipReportBodyDto import TipReportBodyDto
from ._TipReportHeaderDto import TipReportHeaderDto
from ._TipReportFilterDto import TipReportFilterDto
from ._TipReportDetailLineDto import TipReportDetailLineDto

__all__ = [
    "TipReportDto",
    "TipReportFilterDto",
    "TipReportHeaderDto",
    "TipReportBodyDto",
    "TipReportDetailLineDto"]



import os

from logging import getLogger

from cfgtools import read
from msgbus import MBEasyContext
from helper import import_pydevd
from mwhelper import config_logger_parse_file_name
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler

from externals.messagehandlerbuilder import ComponentMessageHandlerBuilder

LOADER_CFG = os.environ["LOADERCFG"]


def main():
    # type: () -> None

    import_pydevd(LOADER_CFG, 9179)

    reader = read(LOADER_CFG)
    service_name = str(reader.find_value("Identification.ComponentName"))
    required_services = "|".join(reader.find_values(service_name + "." + "RequiredServices"))
    persistence_name = reader.find_value(service_name + "." + "RequiredServices").encode("utf-8")

    config_logger_parse_file_name(os.environ["LOG_PATH"], service_name)
    logger = getLogger(__name__)

    mb_context = MBEasyContext(service_name)
    message_bus = MbContextMessageBus(mb_context)
    message_handler_builder = ComponentMessageHandlerBuilder(mb_context, persistence_name)
    message_handler = MbContextMessageHandler(
        message_bus,
        service_name,
        service_name,
        required_services,
        message_handler_builder
    )

    logger.warning("Starting {} Component".format(service_name))
    message_handler.handle_events()


if __name__ == "__main__":
    main()



from application.model.configuration import Configurations
from application.repository.api import APIRepository


class UnlockVoucherInteractor(object):
    def __init__(self, configs, api_repository):
        # type: (Configurations, APIRepository) -> None

        self.logger = configs.logger
        self.configs = configs
        self.api_repository = api_repository

    def unlock_benefit(self, benefit_id):
        # type: (str) -> None

        try:
            self.logger.info("Unlocking benefitId: {}".format(benefit_id))
            self.api_repository.unlock_benefit(benefit_id)
            self.logger.info("Benefit successfully unlocked")
        except Exception as ex:
            self.logger.error("Error at unlock_benefit: {}".format(ex))
            raise ex



class PropertyError(Exception):
    def __init__(self, property_name):
        self.property_name = property_name

    def __str__(self):
        return "Error parsing property {}".format(self.property_name)



class TipReportDetailLineDto(object):
    def __init__(self, operator, table_count, order_count, total_sold_amount, tip_amount):
        # type: (str, int, int, float, float) -> None
        self.__operator = operator
        self.__table_count = table_count
        self.__order_count = order_count
        self.__total_sold_amount = total_sold_amount
        self.__tip_amount = tip_amount

    @property
    def operator(self):
        # type: () -> str
        return str(self.__operator)

    @property
    def table_count(self):
        # type: () -> str
        return str(self.__table_count)

    @property
    def order_count(self):
        # type: () -> str
        return str(self.__order_count)

    @property
    def total_sold_amount(self):
        # type: () -> str
        return str(self.__total_sold_amount)

    @property
    def tip_amount(self):
        # type: () -> str
        return str(self.__tip_amount)



from ._HttpResponseMessage import HttpResponseMessage


class JsonHttpResponseMessage(HttpResponseMessage):
    def __init__(self, status_code, body=None, headers=None):
        if headers is None:
            headers = {}

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json; charset=utf-8"

        super(JsonHttpResponseMessage, self).__init__(status_code, body, headers)



import cfgtools
import os
from copy import copy

import sysactions

from typing import (
    Dict,
    List,
    Optional,
)

from pos_model import (
    SaleItem,
    SimplifiedSaleItem,
)
from pos_util import simplify_sale_items

from basereport import TableReport, ReportColumnDefinition, AlignTypes
from report import Part
from report.command import (
    RepeatCommand,
    AlignCommand,
    FontCommand,
    CutPaperCommand
)
from utils import format_line

from ._ReportBuilder import ReportBuilder
from ..dto import TableReportBodyDto, TableReportOrderDto, TableReportSaleLineDto


class TableReportBody(ReportBuilder):

    def __init__(
        self,
        table_report_body_dto,      # type: TableReportBodyDto
        print_by,                   # type: str
        hide_unpriced_items,        # type: bool
        pre_account_with_discount,  # type: bool
        pos_id,                     # type: str
        filtered_seat="0"           # type: Optional[str]
    ):
        # type: (...) -> None

        self.table_report_body_dto = table_report_body_dto
        self.print_by = print_by
        self.hide_unpriced_items = hide_unpriced_items
        self.pre_account_with_discount = pre_account_with_discount
        self.pos_id = pos_id
        self.filtered_seat = filtered_seat
        self.cut_paper_by_seats = self._get_cut_paper_by_seats_configuration()

    def generate(self):
        # type: () -> List[TableReport]

        lines = []
        table_report_order_dto = self.table_report_body_dto.orders
        table_seats = self.table_report_body_dto.service_seats

        self._configure_font(lines)
        self._generate_items_description(lines)
        self._repeat_of_equals_symbol_line(lines)
        self._generate_order(
            lines=lines,
            table_report_order_dto=table_report_order_dto,
            tip_amount=self.table_report_body_dto.tip,
            detailed_discounts=self.table_report_body_dto.detailed_discounts,
            detailed_payments=self.table_report_body_dto.detailed_payments,
            detailed_tip_excluded=self.table_report_body_dto.detailed_tip_excluded,
            sub_total_amount=self.table_report_body_dto.sub_total_amount,
            total_payment=self.table_report_body_dto.promotional_tenders,
            seats=table_seats,
            discounts_groups=self.table_report_body_dto.discount_groups,
        )

        if self.filtered_seat == "0":
            self._repeat_of_equals_symbol_line(lines)
            self._generate_sub_total_summary(
                lines=lines,
                sub_total_amount=self.table_report_body_dto.sub_total_amount,
                tip_amount=self.table_report_body_dto.tip,
                discounts_amount=self.table_report_body_dto.discount_amount,
                promotional_tenders=self.table_report_body_dto.promotional_tenders,
                discount_groups=self.table_report_body_dto.discount_groups,
            )
            self._generate_total_summary(
                lines,
                self.table_report_body_dto.total_amount
            )
            self._repeat_of_equals_symbol_line(lines)
            if self.table_report_body_dto.service_tenders:
                self._generate_partial_payments_summary(
                    lines=lines,
                    service_tenders=self.table_report_body_dto.service_tenders,
                )
                self._generate_remaining_amount(
                    lines=lines,
                    total_amount=self.table_report_body_dto.total_amount,
                    service_tenders=self.table_report_body_dto.service_tenders,
                )
                self._repeat_of_equals_symbol_line(lines=lines)

            self._generate_total_division_sugestion(
                lines,
                self.table_report_body_dto.total_amount,
                self.table_report_body_dto.service_seats
            )
            self._generate_customer_doc_field(lines)
            self._repeat_of_under_symbol_line(lines)
        elif self.table_report_body_dto.service_tenders:
            self._repeat_of_equals_symbol_line(lines=lines)
            self._generate_warning_partial_payments(lines=lines)
            self._repeat_of_equals_symbol_line(lines=lines)
        self._a_new_empty_line(lines)
        self._generate_footer(lines)

        return lines

    def _generate_order(
        self,
        lines,                      # type: List
        table_report_order_dto,     # type: List[TableReportOrderDto]
        tip_amount,                 # type: float
        detailed_discounts,         # type: Dict
        detailed_payments,          # type: Dict
        detailed_tip_excluded,      # type: Dict
        sub_total_amount,           # type: float
        total_payment,              # type: float
        seats,                      # type: int
        discounts_groups,           # type: Dict
    ):
        # type: (...) -> TableReportBody

        grouped_sale_lines = self._get_grouped_sale_lines(table_report_order_dto, seats)
        fraction_line_numbers = []
        for _, sale_lines in grouped_sale_lines.items():
            fraction_line_numbers.extend(self._get_fraction_sale_line_identifiers(sales_line_list=sale_lines))

        if self.print_by in ["SEATS", "TAB"] and len(grouped_sale_lines) > 1:
            simplified_sale_lines = {}
            for seat, sale_lines in grouped_sale_lines.items():
                simplified_sale_lines[seat] = _simplify_table_report_sale_lines(
                    sale_lines,
                    hide_unpriced_items=self.hide_unpriced_items,
                    fraction_line_numbers=fraction_line_numbers,
                )
            lines.extend(
                self._generate_orders_split_by_seats_from_sale_lines_with_tab_on_modifiers(
                    grouped_sale_lines=simplified_sale_lines,
                    tip_amount=tip_amount,
                    detailed_discounts=detailed_discounts,
                    detailed_payments=detailed_payments,
                    detailed_tip_excluded=detailed_tip_excluded,
                    sub_total_amount=sub_total_amount,
                    total_payment=total_payment,
                    printed_by=self.print_by,
                    filtered_seat=self.filtered_seat,
                    discounts_groups=discounts_groups,
                )
            )
        else:
            all_orders_sale_lines = []
            for order in table_report_order_dto:
                all_orders_sale_lines.extend(order.sales_line_list)
            simplified_sale_lines = _simplify_table_report_sale_lines(
                all_orders_sale_lines,
                hide_unpriced_items=self.hide_unpriced_items,
                fraction_line_numbers=fraction_line_numbers,
            )
            lines.extend(self._generate_order_without_split_seats(sale_lines=simplified_sale_lines))
            lines.append(a_new_line())

        return self

    @staticmethod
    def _get_parent_table_seat(sales_line_list, line_number):
        seat = [x.seat for x in sales_line_list if x.level == 0 and x.line_number == line_number]
        if len(seat) > 0:
            return seat[0]
        return 0

    def _get_table_seat(self, sales_line_list, sale_line):
        if sale_line.level == 0:
            seat = sale_line.seat
        else:
            seat = self._get_parent_table_seat(sales_line_list, sale_line.line_number)

        if seat == 0:
            seat = "$TABLE"

        return seat

    def _get_grouped_sale_lines(self, table_report_order_dto, seats):
        grouped_sale_lines = {}
        for i in range(seats):
            grouped_sale_lines[i+1] = []

        for order in table_report_order_dto:
            for sale_line in order.sales_line_list:
                if sale_line.item_type == "COUPON":
                    continue

                if sale_line.item_type == "COMBO" and sale_line.qty == 0:
                    continue

                seat = self._get_table_seat(order.sales_line_list, sale_line)

                if seat not in grouped_sale_lines:
                    grouped_sale_lines[seat] = [sale_line]
                else:
                    grouped_sale_lines[seat].append(sale_line)
        return grouped_sale_lines

    @staticmethod
    def _get_cut_paper_by_seats_configuration():
        config = cfgtools.read(os.environ["LOADERCFG"])
        cut_paper_config = (config.find_value("TableReport.CutPaperBySeat") or "false").lower() == "true"

        return cut_paper_config

    def _generate_order_without_split_seats(
        self,
        sale_lines,     # type: List[TableReportSaleLineDto]
    ):
        # type: (...) -> List[TableReport]

        order_items_lines = self._build_product_lines(sales_line_list=sale_lines)
        order_lines = TableReport(
            lines=order_items_lines,
            column_definitions=a_sub_total_summary_column_definition(),
        )
        return [order_lines]

    def _generate_orders_split_by_seats_from_sale_lines_with_tab_on_modifiers(
        self,
        grouped_sale_lines,     # type: Dict
        tip_amount,             # type: float
        detailed_discounts,     # type: Dict
        detailed_payments,      # type: Dict
        detailed_tip_excluded,  # type: Dict
        sub_total_amount,       # type: float
        total_payment,          # type: float
        printed_by,             # type: str
        filtered_seat='0',      # type: Optional[str]
        discounts_groups=None,  # type: Optional[Dict]
    ):
        # type: (...) -> List

        if printed_by == "TAB":
            grouped_sale_lines = {k: v for k, v in grouped_sale_lines.items() if v}

        tip_excluded_amount = sum(detailed_tip_excluded.values())
        tip_percentage = self._get_percentage(tip_amount, sub_total_amount - tip_excluded_amount)
        seat_quantities = len(grouped_sale_lines.keys()) - (1 if "$TABLE" in grouped_sale_lines else 0)

        not_mapped_payment = total_payment - sum(detailed_payments.values())

        table_tip_excluded = detailed_tip_excluded[0] if 0 in detailed_tip_excluded.keys() else 0.0
        table_discount = detailed_discounts[0] if 0 in detailed_discounts.keys() else 0.0
        table_payment = detailed_payments[0] if 0 in detailed_payments.keys() else 0.0
        table_sale_line = grouped_sale_lines["$TABLE"] if "$TABLE" in grouped_sale_lines else {}
        table_sub_total = 0.0 + sum([float(x.price) if x.price and x.qty != 0 else 0.0 for x in table_sale_line])
        table_tip = ((tip_percentage * (table_sub_total - table_tip_excluded)) / 100)
        table_per_seat = round(table_sub_total / seat_quantities, 2)
        table_discount_per_seat = table_discount / seat_quantities
        table_payment_per_seat = table_payment / seat_quantities
        table_tip_per_seat = table_tip / seat_quantities

        sub_total_seats_amount = sub_total_amount - table_sub_total or 1

        lines = []
        sorted_grouped_sale_lines = sorted(grouped_sale_lines, key=lambda g: (1, g) if isinstance(g, int) else (0, g))
        for index, seat in enumerate(sorted_grouped_sale_lines):
            if str(seat) != filtered_seat and filtered_seat != '0':
                continue
            seat_lines = []
            add_to_report = True
            if filtered_seat == '0':
                if self.cut_paper_by_seats:
                    seat_lines.append(Part(commands=[CutPaperCommand()]))
                elif index != 0:
                    seat_lines.append(Part(commands=[RepeatCommand(38, "-")]))
                    seat_lines.append(a_new_line())

            is_table = seat == "$TABLE"
            name_by = "$SEAT_NUMBER" if printed_by == "SEATS" else "$TAB_NUMBER"
            seat_descr = seat if is_table else "{} {}".format(name_by, seat)
            seat_sub_total = sum([float(x.price) if x.price and x.qty != 0 else 0.0 for x in grouped_sale_lines[seat]])
            discounts_groups_seat = copy(discounts_groups)
            seat_tip = 0
            seat_total = 0
            seat_payment = 0
            seat_discount = 0
            if not is_table:
                for key in discounts_groups_seat.keys():
                    value = discounts_groups_seat[key]
                    seat_discount = value * (seat_sub_total / sub_total_seats_amount)
                    seat_discount += detailed_discounts[seat] if seat in detailed_discounts.keys() else 0.0
                    seat_discount = round(seat_discount + table_discount_per_seat, 2)
                    discounts_groups_seat[key] = seat_discount

                seat_tip_excluded = detailed_tip_excluded[seat] if seat in detailed_tip_excluded.keys() else 0.0
                seat_payment = not_mapped_payment * (seat_sub_total / sub_total_seats_amount)
                seat_payment += detailed_payments[seat] if seat in detailed_payments.keys() else 0.0
                seat_payment = round(seat_payment + table_payment_per_seat, 2)
                seat_tip = (tip_percentage * (seat_sub_total - seat_tip_excluded)) / 100
                seat_tip = round(seat_tip + table_tip_per_seat, 2)
                seat_total = seat_sub_total + table_per_seat - seat_discount + seat_tip - seat_payment

            seat_title = [
                [seat_descr, ""]
            ]
            seat_lines.append(TableReport(lines=seat_title, column_definitions=a_total_summary_column_definitions()))
            seat_lines.append(a_new_line())

            order_items = self._build_product_lines(sales_line_list=grouped_sale_lines[seat])
            order_lines = TableReport(
                lines=order_items,
                column_definitions=a_sub_total_summary_column_definition(),
            )
            seat_lines.append(order_lines)
            seat_lines.append(a_new_line())

            subtotal = [["$SUBTOTAL", "#NUMBER({})".format(seat_sub_total)]]
            seat_lines.append(
                TableReport(
                    lines=subtotal,
                    column_definitions=a_total_detail_summary_column_definition(),
                )
            )

            seat_details = []
            has_details = False
            if table_per_seat > 0 and is_table:
                seat_details.append(["$DIVISION_FOR_SEATS", "#NUMBER({})".format(table_per_seat)])
            if table_per_seat > 0 and not is_table:
                seat_details.append(["$DIVISION_FROM_TABLE", "#NUMBER({})".format(table_per_seat)])
            seat_lines.append(
                TableReport(
                    lines=seat_details,
                    column_definitions=a_total_detail_summary_column_definition(),
                )
            )

            seat_details = []
            if seat_discount > 0:
                self._generate_discount_field(
                    report_lines=seat_lines,
                    discount_list=seat_details,
                    discount_amount=seat_discount,
                    discount_groups=discounts_groups_seat,
                )

            if seat_payment > 0 and not is_table:
                seat_details.append(["$PROMOTIONAL_TENDERS", "#NUMBER({})".format(-seat_payment)])
            if seat_tip > 0:
                seat_details.append(["$TIP", "#NUMBER({})".format(seat_tip)])

            if len(seat_details) > 0:
                has_details = True

            if not is_table:
                seat_details.append(["$TOTAL", "#NUMBER({})".format(seat_total)])
            if not has_details and len(order_lines.lines) == 0:
                add_to_report = False
            seat_lines.append(
                TableReport(
                    lines=seat_details,
                    column_definitions=a_total_detail_summary_column_definition(),
                )
            )

            if filtered_seat == '0' and self.cut_paper_by_seats and index == (len(grouped_sale_lines) - 1):
                seat_lines.append(Part(commands=[CutPaperCommand()]))

            if add_to_report:
                lines.extend(seat_lines)

        return lines

    @staticmethod
    def _get_percentage(amount, sub_total_amount):
        if amount > 0:
            percentage = round((amount * 100) / sub_total_amount, 2)
        else:
            percentage = 0
        return percentage

    @staticmethod
    def _build_product_lines(
        sales_line_list,    # type: List[TableReportSaleLineDto]
    ):
        # type: (...) -> List

        order_items_lines = []
        for item in sales_line_list:
            item_service_period = "[{}]".format(item.service_period) if item.service_period else " "

            item_price = item.price or 0
            item_price = "#NUMBER({})".format(item_price) if item_price else ""

            comment_description = ""
            if item.comment and item.comment.startswith("["):
                comment_description = "$" + item.comment.replace("[", "").replace("]", "").upper() + " "

            product_description = comment_description + item.product_name
            fraction = item.fraction_parent
            if fraction:
                numerator = int(round(item.qty * fraction))
                qty_description = " {}/{}".format(numerator, fraction) if numerator != fraction else "1.00"
            else:
                qty_description = "{:.2f}".format(item.qty)
            qty_description = qty_description.rjust(5)
            indentation = " " * (item.level + 1)
            description = u"{}{}".format(indentation, product_description).ljust(30)[:30]
            item_price = item_price.rjust(38 - len(description) + len(qty_description))

            order_items_lines.append([qty_description, description, item_service_period, item_price])

        return order_items_lines

    @staticmethod
    def _get_fraction_sale_line_identifiers(
        sales_line_list,    # type: List[TableReportSaleLineDto]
    ):
        # type: (...) -> List

        fraction_sale_lines_list = []
        for sale_line in sales_line_list:
            if sale_line.fraction_parent:
                line_number_identifier = "{}-{}".format(sale_line.order_id, sale_line.line_number)
                fraction_sale_lines_list.append(line_number_identifier)

        return fraction_sale_lines_list

    def _generate_sub_total_summary(
        self,
        lines,                  # type: List
        sub_total_amount,       # type: float
        tip_amount,             # type: float
        discounts_amount,       # type: float
        promotional_tenders,    # type: float
        discount_groups,        # type: dict
    ):
        # type: (...) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()

        sub_total_lines = [["$SUBTOTAL", "#NUMBER({})".format(sub_total_amount)]]

        lines.extend([
            TableReport(
                lines=sub_total_lines,
                column_definitions=total_summary_column_definitions,
            )
        ])

        if discounts_amount > 0:
            self._generate_discount_field(
                report_lines=lines,
                discount_list=sub_total_lines,
                discount_amount=discounts_amount,
                discount_groups=discount_groups,
            )

        sub_total_lines = []
        if promotional_tenders > 0:
            sub_total_lines.append(["$PROMOTIONAL_TENDERS", "#NUMBER({})".format(-promotional_tenders)])
        if tip_amount > 0:
            sub_total_lines.append(["$SUGGESTED_TIP", "#NUMBER({})".format(tip_amount)])

        lines.extend([
            TableReport(
                lines=sub_total_lines,
                column_definitions=total_summary_column_definitions,
            ),
            a_new_line(),
        ])
        return self

    def _generate_discount_field(
        self,
        report_lines,       # type: List
        discount_list,      # type: List
        discount_amount,    # type: float
        discount_groups,    # type: Dict
    ):
        # type: (...) -> None

        total_summary_column_definitions = a_total_summary_column_definitions()
        discount_lines = []
        has_groups = len(discount_groups) > 1
        if has_groups:
            report_lines.extend([a_new_line()])
            report_lines.extend([
                Part(
                    text="$DISCOUNTS",
                    commands=[AlignCommand(AlignCommand.Alignment.left)],
                ),
                a_new_line(),
            ])
            self._parse_discounts_group(
                discounts_list=discount_lines,
                discount_groups=discount_groups,
            )
            report_lines.extend([
                TableReport(
                    lines=discount_lines,
                    column_definitions=total_summary_column_definitions,
                ),
                a_new_line(),
            ])
        else:
            discount_list.append(["$DISCOUNTS", "#NUMBER({})".format(-discount_amount)])

    @staticmethod
    def _parse_discounts_group(
        discounts_list,         # type: List
        discount_groups=None,   # type: Dict
    ):
        # type: (...) -> None

        for discount in discount_groups:
            value = float(discount_groups.get(discount))
            discounts_list.append(["{}".format(discount), "#NUMBER({})".format(-value)])

    def _generate_total_summary(self, lines, total_amount):
        # type: (List, float) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()

        sub_total_lines = [
            ["$TOTAL", "#NUMBER({})".format(total_amount)]
        ]
        lines.extend([TableReport(lines=sub_total_lines, column_definitions=total_summary_column_definitions)])
        return self

    def _generate_partial_payments_summary(
        self,
        lines,              # type: List
        service_tenders,    # type: List[dict]
    ):
        # type: (...) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()
        payments = []
        for tender in service_tenders:
            tender_amount = tender.get('tender_amount')
            payments.append([tender.get('tender_descr'), "#NUMBER({})".format(tender_amount)])

        total_paid = sum([float(tender.get('tender_amount')) for tender in service_tenders])
        payments.append(["$PAID", "#NUMBER({})".format(total_paid)])

        lines.extend(
            [
                TableReport(
                    lines=payments,
                    column_definitions=total_summary_column_definitions,
                ),
                a_new_line()
            ]
        )

        return self

    def _generate_remaining_amount(
        self,
        lines,              # type: List
        total_amount,       # type: float
        service_tenders,    # type: List[dict]
    ):
        # type: (...) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()

        total_paid = sum([float(service_tender.get('tender_amount')) for service_tender in service_tenders])
        to_pay = [["$TO_PAY", "#NUMBER({})".format(total_amount - total_paid)]]
        lines.extend([
            TableReport(
                lines=to_pay,
                column_definitions=total_summary_column_definitions,
            )
        ])

        return self

    def _generate_warning_partial_payments(
        self,
        lines,              # type: List
        columns_width=38,   # type: int
    ):
        # type: (...) -> TableReportBody

        model = sysactions.get_model(self.pos_id)
        message = sysactions.translate_message(model, "PAYMENTS_NOT_DISTRIBUTED")

        message_size = len(message)
        begin = 0
        while begin < message_size:
            lines.extend([
                Part(
                    text=message[begin:begin + columns_width],
                    commands=[AlignCommand(AlignCommand.Alignment.left)],
                ),
                a_new_line(),
            ])
            begin += columns_width

        return self

    def _generate_total_division_sugestion(self, lines, total_amount, service_seats):
        # type: (List, float, int) -> TableReportBody

        if service_seats > 1:
            division_amount = total_amount / service_seats

            lines.extend([
                Part("$EQUAL_DIVISION_FOR_ALL_SEATS", [AlignCommand(AlignCommand.Alignment.left)]), a_new_line()
            ])

            sub_total_lines = [
                ["$DIVISION_SUGESTION_FOR {}".format(service_seats), "#NUMBER({})".format(division_amount)]]

            lines.extend(
                [TableReport(lines=sub_total_lines, column_definitions=a_total_division_sugestion_definitions()),
                 a_new_line()
                 ])
        return self

    def _generate_customer_doc_field(self, lines):
        # type: (List) -> TableReportBody

        total_division_sugestion_definitions = a_customer_doc_field()

        sub_total_lines = [["$TYPE_YOUR_DOC_NUM"]]
        lines.extend(
            [TableReport(lines=sub_total_lines, column_definitions=total_division_sugestion_definitions),
             a_new_line()
             ])
        return self

    def _generate_footer(self, lines):
        # type: (List) -> TableReportBody
        lines.extend([
            Part("$THANK_YOU_FOR_YOUR_PREFERENCE", [AlignCommand(AlignCommand.Alignment.center)]), a_new_line(),
            Part("$SERVICE_TAX_IS_NOT_MANDATORY", [AlignCommand(AlignCommand.Alignment.center)]), a_new_line()
        ])

        if not self.pre_account_with_discount:
            model = sysactions.get_model(self.pos_id)
            message = sysactions.translate_message(model, "WARNING_REPORT_ITEMS_CONFERENCE")
            self._a_new_empty_line(lines)
            lines.extend(format_line('', message, columns=38, definition=a_information_footer()))

        self._a_new_empty_line(lines)
        return self

    def _generate_items_description(self, lines):
        # type: (List[TableReport]) -> TableReportBody

        items_description = TableReport(
            lines=[["$QTY", "$ITEM", "#CURRENCY_SYMBOL()"]],
            column_definitions=a_item_column_definition())

        lines.append(items_description)
        return self

    def _configure_font(self, lines):
        # type: (List[TableReport]) -> TableReportBody
        lines.append(Part(commands=[FontCommand(FontCommand.Font.A)]))

        return self

    def _repeat_of_equals_symbol_line(self, lines):
        # type: (List) -> TableReportBody
        repeat_symbol = [Part(commands=[RepeatCommand(38, "=")]), a_new_line()]
        lines.extend(repeat_symbol)

        return self

    def _repeat_of_under_symbol_line(self, lines):
        # type: (List) -> TableReportBody
        repeat_symbol = [Part(commands=[RepeatCommand(38, "_")]), a_new_line()]
        lines.extend(repeat_symbol)

        return self

    def _a_new_empty_line(self, lines):
        # type: (List) -> TableReportBody
        lines.extend([a_new_line()])

        return self


def _simplify_table_report_sale_lines(
    sale_lines,                     # type: List[TableReportSaleLineDto]
    hide_unpriced_items=False,      # type: bool
    fraction_line_numbers=None,     # type: Optional[List]
):
    # type: (...) -> List[TableReportSaleLineDto]

    items = [
        _create_simplified_sale_item(
            sale_line=sale_line,
            fraction_line_numbers=fraction_line_numbers,
        ) for sale_line in sale_lines
    ]
    simplified_sale_items = simplify_sale_items(
        items,
        hide_unpriced_items=hide_unpriced_items,
    )
    result_sale_lines = []

    for item in simplified_sale_items:
        sale_line = item.original_reference  # type: TableReportSaleLineDto
        sale_line.level = item.sale_item.level
        sale_line.price = item.sale_item.item_price
        sale_line.qty = item.sale_item.quantity
        if item.sale_item.default_quantity is not None:
            sale_line.default_qty = item.sale_item.default_quantity
        result_sale_lines.append(sale_line)

    return result_sale_lines


def _create_simplified_sale_item(
    sale_line,                      # type: TableReportSaleLineDto
    fraction_line_numbers=None,     # type: Optional[Dict]
):
    # type: (...) -> SimplifiedSaleItem

    sale_item = SaleItem(
        line_number=sale_line.line_number,
        level=sale_line.level,
        part_code=sale_line.part_code,
        product_name=sale_line.product_name,
        item_id=sale_line.item_id,
        item_price=sale_line.price,
        quantity=sale_line.qty,
    )
    sale_item.overwritten_unit_price = sale_line.overwritten_unit_price
    sale_item.seat = sale_line.seat
    sale_item.comment = sale_line.comment
    sale_item.service_period = sale_line.service_period
    sale_item.default_quantity = sale_line.default_qty
    sale_item.type = sale_line.item_type
    sale_line_identifier = "{}-{}".format(sale_line.order_id, sale_line.line_number)

    item = SimplifiedSaleItem(
        sale_item,
        order_id=sale_line.order_id,
        original_reference=sale_line,
        has_fraction=sale_line_identifier in fraction_line_numbers
    )
    return item


def a_new_line():
    return Part("\n")


def a_item_column_definition():
    return [
        ReportColumnDefinition(width=5,
                               before_text="",
                               after_text="",
                               after_fill_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text="  ",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=8,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_sub_total_summary_column_definition():
    return [
        ReportColumnDefinition(width=5,
                               before_text="",
                               after_text="",
                               after_fill_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=22,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=4,
                               fill_with_char=" ",
                               align=AlignTypes.RIGHT),
        ReportColumnDefinition(width=7,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_detail_summary_column_definition():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.RIGHT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_summary_column_definitions():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_division_sugestion_definitions():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_customer_doc_field():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.CENTER),
    ]


def a_information_footer():
    return [
        ReportColumnDefinition(width=38,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER),
    ]



from helper import ExtendedEnum
from typing import Any


class ValidationError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class AuthValidationErrors(Exception):
    def __init__(self, ex):
        # type: (ExtendedEnum) -> None
        self.error_type = ex.name
        self.error_message = ex.value


class ExternalServiceErrors(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class OperationErrors(Exception):
    def __init__(self, ex):
        # type: (ExtendedEnum) -> None
        self.error_type = ex.name
        self.error_message = ex.value


class BadRequestError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class NotFoundError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class ClientCustomError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class InUseError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value


class NotAvailableError(Exception):
    def __init__(
            self,
            ex,     # type: ExtendedEnum
            *args   # type: Any
    ):
        # type: (...) -> None

        self.error_type = ex.name
        if args:
            self.error_message = ex.value.format(*args)
        else:
            self.error_message = ex.value



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

import six

from cryptography import utils
from cryptography.exceptions import (
    AlreadyFinalized, InvalidKey, UnsupportedAlgorithm, _Reasons
)
from cryptography.hazmat.backends.interfaces import HMACBackend
from cryptography.hazmat.primitives import constant_time, hmac
from cryptography.hazmat.primitives.kdf import KeyDerivationFunction


@utils.register_interface(KeyDerivationFunction)
class HKDF(object):
    def __init__(self, algorithm, length, salt, info, backend):
        if not isinstance(backend, HMACBackend):
            raise UnsupportedAlgorithm(
                "Backend object does not implement HMACBackend.",
                _Reasons.BACKEND_MISSING_INTERFACE
            )

        self._algorithm = algorithm

        if salt is None:
            salt = b"\x00" * self._algorithm.digest_size
        else:
            utils._check_bytes("salt", salt)

        self._salt = salt

        self._backend = backend

        self._hkdf_expand = HKDFExpand(self._algorithm, length, info, backend)

    def _extract(self, key_material):
        h = hmac.HMAC(self._salt, self._algorithm, backend=self._backend)
        h.update(key_material)
        return h.finalize()

    def derive(self, key_material):
        utils._check_byteslike("key_material", key_material)
        return self._hkdf_expand.derive(self._extract(key_material))

    def verify(self, key_material, expected_key):
        if not constant_time.bytes_eq(self.derive(key_material), expected_key):
            raise InvalidKey


@utils.register_interface(KeyDerivationFunction)
class HKDFExpand(object):
    def __init__(self, algorithm, length, info, backend):
        if not isinstance(backend, HMACBackend):
            raise UnsupportedAlgorithm(
                "Backend object does not implement HMACBackend.",
                _Reasons.BACKEND_MISSING_INTERFACE
            )

        self._algorithm = algorithm

        self._backend = backend

        max_length = 255 * algorithm.digest_size

        if length > max_length:
            raise ValueError(
                "Can not derive keys larger than {} octets.".format(
                    max_length
                ))

        self._length = length

        if info is None:
            info = b""
        else:
            utils._check_bytes("info", info)

        self._info = info

        self._used = False

    def _expand(self, key_material):
        output = [b""]
        counter = 1

        while self._algorithm.digest_size * (len(output) - 1) < self._length:
            h = hmac.HMAC(key_material, self._algorithm, backend=self._backend)
            h.update(output[-1])
            h.update(self._info)
            h.update(six.int2byte(counter))
            output.append(h.finalize())
            counter += 1

        return b"".join(output)[:self._length]

    def derive(self, key_material):
        utils._check_byteslike("key_material", key_material)
        if self._used:
            raise AlreadyFinalized

        self._used = True
        return self._expand(key_material)

    def verify(self, key_material, expected_key):
        if not constant_time.bytes_eq(self.derive(key_material), expected_key):
            raise InvalidKey



from datetime import datetime
from typing import List

class SaleChallengeItem(object):

    def __init__(self, name, start_date, end_date, products):
        # type: (int, datetime, datetime, List[int]) -> None

        self.name = name
        self.start_date = start_date
        self.end_date = end_date
        self.products = products



# -*- coding: utf-8 -*-

from logging import Logger

import sysactions
from msgbus import MBEasyContext

from devicemanager.interactor import GetUsedServiceInteractor
from devicemanager.model import (
    Configs,
    ListenedTokens,
)
from devicemanager.processor import GetUsedServiceProcessor
from devicemanager.util import CacheManager
from messagebus import (
    Event,
    Message,
    MessageBus,
)
from messagehandler import (
    MessageHandler,
    MessageHandlerBuilder,
)
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from typing import Any


class DeviceManagerMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(
        self,
        configs,            # type: Configs
        message_bus,        # type: MessageBus
        mb_context,         # type: MBEasyContext
        logger,             # type: Logger
        input_logger,       # type: InputLoggerService
        callback_logger,    # type: Logger
    ):
        # type: (...) -> None

        self.configs = configs
        self.message_bus = message_bus
        self.mb_context = mb_context
        self.logger = logger
        self.input_logger = input_logger
        self.callback_logger = callback_logger
        self.cache_manager = None
        sysactions.mbcontext = message_bus.mbcontext

    def build_singletons(
        self,
    ):
        # type: (...) -> None

        self.cache_manager = CacheManager(sector_config=self.configs.sector_config)

    def build_message_handler(
        self,
        message,    # type: Message
        event,      # type: Event
    ):
        # type: (...) -> MessageHandler

        get_used_service_interactor = GetUsedServiceInteractor(cache_manager=self.cache_manager)
        get_used_service_processor = GetUsedServiceProcessor(get_used_service_interactor=get_used_service_interactor)

        message_handler = MessageProcessorMessageHandler(
            event_processors=None,
            message_processors={
                ListenedTokens.TK_GET_USED_SERVICE.value: get_used_service_processor
            },
            executor_factory=DefaultMessageProcessorExecutorFactory(
                [
                    LoggingProcessorCallback(self.callback_logger)
                ]
            ),
        )

        return message_handler

    def destroy_singletons(
        self,
    ):
        # type: (...) -> None

        pass

    def destroy_message_handler(
        self,
        message_handler,    # type: Any
    ):
        # type: (...) -> None

        pass



# coding=utf-8
import json
import os

import cfgtools
import sysactions

from bustoken import TK_REMOTE_ORDER_GET_LOGISTIC_PARTNERS, TK_REMOTE_ORDER_SEARCH_LOGISTIC
from msgbus import TK_SYS_ACK
from sysactions import action, show_listbox, translate_message, StopAction, show_messagebox, get_model

from .. import logger, mb_context
from application.interactor import UserInteractor
from model.customexception import LogisticPartnerNotFound, LogisticSearchError

logistic_partners = {}


@action
def select_logistic_partner(pos_id, order_id):
    # type: (str, str) -> None

    select_partner = _ask_for_logistic_partner(pos_id)
    if select_partner:
        _search_logistic_partner(order_id, pos_id, select_partner)


def _search_logistic_partner(order_id, pos_id, select_partner):
    params = [str(select_partner["id"]), order_id]

    if _should_ask_for_deliveryman_name():
        if select_partner["id"] == 0:
            own_delivery_men = UserInteractor().get_own_delivery_men()
            if own_delivery_men:
                index = show_listbox(pos_id, own_delivery_men, message="$DELIVERYMEN")
                if index is None:
                    raise StopAction()

                deliveryman_data = own_delivery_men[index]
                params.append(deliveryman_data)
            else:
                deliveryman_data = sysactions.show_keyboard(pos_id, "$DELIVERY_MAN_NAME")
                if deliveryman_data is not None and deliveryman_data != "":
                    params.append(deliveryman_data)
                else:
                    show_messagebox(pos_id, "$INSERT_DELIVERY_MAN_NAME")
                    return

    data = json.dumps("\0".join(params))
    msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_SEARCH_LOGISTIC, data=data, timeout=10000000)
    if msg.token == TK_SYS_ACK:
        show_messagebox(pos_id, "$SEARCHING_LOGISTIC_PARTNER")
    else:
        raise LogisticSearchError()


def _ask_for_logistic_partner(
    pos_id,     # type: str
):
    # type: (...) -> int

    try:
        partners = _get_logistic_partners()
    except LogisticPartnerNotFound as _:
        show_messagebox(
            posid=pos_id,
            message="$LOGISTIC_NOT_FOUND",
        )
        raise StopAction()

    listbox_options = map(lambda p: p["name"], partners)
    index = show_listbox(pos_id, listbox_options, message="$DELIVERYMEN")
    if index is None:
        raise StopAction()

    select_partner = partners[index]
    return select_partner


def _get_logistic_partners():
    global logistic_partners

    if not logistic_partners:
        logger.debug("Getting Logistic Partners from Remote Order")
        msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_GET_LOGISTIC_PARTNERS, timeout=10000000)
        if msg.token == TK_SYS_ACK:
            logger.debug("Logistic Partners from Remote Order: {}".format(msg.data))
            logistic_partners = json.loads(msg.data)
        else:
            logger.debug("Error getting Logistic Partners from Remote Order")

    if not logistic_partners:
        raise LogisticPartnerNotFound()

    return logistic_partners


def _should_ask_for_deliveryman_name():
    config = cfgtools.read(os.environ["LOADERCFG"])
    value = config.find_value("Customizations.DeliveryConfigurations.AskForDeliveryManName", False)
    value = str(value).upper() == 'TRUE'
    return value



from xml.etree import cElementTree as eTree
from abc import (
    ABCMeta,
    abstractmethod,
)

from posot import OrderTaker

from messagebus import Message
from typing import Optional


class ElectronicPaymentServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def finish_transaction(
        self,
        model,                  # type: eTree
        pos_ot,                 # type: OrderTaker
        electronic_type_id,     # type: int
        pos_id,                 # type: str
        status,                 # type: str
        order_id="",            # type: Optional[str]
    ):
        # type: (...) -> Optional[Message]

        raise NotImplementedError



class TokenCreator(object):
    #
    # M E S S A G E   T O K E N S
    #
    # TOKEN FORMAT (32 bits): 0xPGGXNNNN
    #    P  - message priority
    #    G  - message group
    #    X  - not used
    #    N  - message number
    #
    @staticmethod
    def create_token(priority, message_group, number):
        # type: (str, str, str) -> int
        while len(number) < 4:
            number = "0" + number

        while len(message_group) < 2:
            message_group = "0" + message_group

        return int("0x" + priority + message_group + "0" + number, 16)



from production.box import AllServedItemsBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class AllServedItemsBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(AllServedItemsBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> AllServedItemsBoxBuilder
        return AllServedItemsBox(config.name, self.get_logger(config))



from msgbus import FM_PARAM, TK_SYS_ACK, TK_SYS_NAK
from bustoken import create_token, MSGPRT_LOW

from utilfunctions import json_serialize

from customer_model import AddCustomerRequest

from .. import mb_context

TK_CUSTOMER_ADD_CUSTOMER = create_token(MSGPRT_LOW, "96", "1")
TK_CUSTOMER_GET_CUSTOMER = create_token(MSGPRT_LOW, "96", "2")
TK_CUSTOMER_UPDATE_CUSTOMER = create_token(MSGPRT_LOW, "96", "3")
TK_CUSTOMER_DELETE_CUSTOMER = create_token(MSGPRT_LOW, "96", "4")


class Customer(object):
    @staticmethod
    def add_customer(add_customer_request):
        # type: (AddCustomerRequest) -> None

        data = json_serialize(add_customer_request)
        response = mb_context.MB_EasySendMessage("Customer", TK_CUSTOMER_ADD_CUSTOMER, FM_PARAM, data)

        if response.token == TK_SYS_NAK:
            raise Exception("Error")

    @staticmethod
    def get_customer(phone):
        # type: (str) -> str

        response = mb_context.MB_EasySendMessage("Customer", TK_CUSTOMER_GET_CUSTOMER, FM_PARAM, str(phone))

        if response.token == TK_SYS_ACK:
            return response.data

        return ''

    @staticmethod
    def update_customer(add_customer_request):
        # type: (AddCustomerRequest) -> None

        data = json_serialize(add_customer_request)
        response = mb_context.MB_EasySendMessage("Customer", TK_CUSTOMER_UPDATE_CUSTOMER, FM_PARAM, data)

        if response.token == TK_SYS_NAK:
            raise Exception("Error")



class NfceNamespaces(object):
    nfe = "http://www.portalfiscal.inf.br/nfe"
    NAMESPACE_SOAP = "http://www.w3.org/2003/05/soap-envelope"
    NAMESPACE_AUTORIZACAO = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao"
    NAMESPACE_AUTORIZACAO_3 = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao3"
    NAMESPACE_RET_AUTORIZACAO = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeRetAutorizacao"
    NAMESPACE_CONSULTA_SITUACAO = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeConsulta2"


# -*- coding: utf-8 -*-

import base64
import json
from xml.etree import ElementTree as eTree
import sysactions
from common import PrinterCommon
from fiscalpersistence import Tender
from msgbus import MBEasyContext
from typing import (
    List,
    Optional,
    Dict,
)

from _AposA8PrintLine import AposA8PrintLine
from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8Constants import AposA8Constants

from printer import FiscalPrinter
from printer_wrapper import insert_space
from clichehelper import ClicheHelper
from _AposA8Helpers import (
    append_and_count_sale_items,
    append_coupon_fiscal_parameters,
    append_customer_name,
)
from _AposA8FiscalPrinter import AposA8FiscalPrinter


class SatAposA8FiscalPrinter(AposA8FiscalPrinter):
    def __init__(
        self,
        mb_context,                     # type: MBEasyContext
        print_customer_name,            # type: bool
        cliche_helper,                  # type: ClicheHelper
        customer_text_field,            # type: str
        printer_common,                 # type: PrinterCommon
        coupon_fiscal_parameters,       # type: Dict[str, bool]
        font_size_withdrawal_panel,     # type: str
    ):
        # type: (...) -> None

        super(SatAposA8FiscalPrinter, self).__init__(
            mb_context,
            print_customer_name,
            cliche_helper,
            customer_text_field,
            coupon_fiscal_parameters,
            font_size_withdrawal_panel,
        )

        self.customer_text_field = customer_text_field
        self.printer_common = printer_common
        self.cliche_helper = cliche_helper
        self.font_size_withdrawal_panel = font_size_withdrawal_panel

    def print_fiscal(self, posid, order, tenders, fiscal_request, fiscal_response, total_taxes, customer_name, url_sefaz=""):
        # type: (str, eTree.Element, list, eTree.Element, eTree.Element, dict, str, str) -> str
        xml = base64.b64decode(fiscal_request.text.strip())
        parsed_xml = eTree.XML(xml)
        n_c_fe = parsed_xml.find("infCFe/ide/nCFe").text
        model = sysactions.get_model(posid)

        report = self._fiscal_report(order, tenders, fiscal_response, n_c_fe, customer_name, model)

        qr_code = fiscal_response.find("QRCode").text.strip()
        sat_key = fiscal_response.find("satkey").text.strip()
        report.append(AposA8PrintLine(AposA8Align.CENTER, sat_key, AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, '', AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, qr_code, AposA8CommandType.PRINT_QR_CODE))

        text = "Consulte o QRCode pelo Aplicativo"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))
        text = "De olho na nota disponivel"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))
        text = "na AppStore (Apple) e Play Store (Android)"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        tax_text = "%s %7.2f" % ("Valor aproximado dos tributos deste cupom", (
                    float(total_taxes['nacionalfederal']['value']) + float(total_taxes['estadual']['value'])))
        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      tax_text,
                                      AposA8CommandType.PRINT_TEXT))

        federal_tax = "Fed = R$%7.2f (%.2f%%), Est = R$%7.2f (%.2f%%)" % (
            total_taxes['nacionalfederal']['value'], total_taxes['nacionalfederal']['percent'],
            total_taxes['estadual']['value'], total_taxes['estadual']['percent'])

        report.append(AposA8PrintLine(AposA8Align.CENTER, federal_tax,  AposA8CommandType.PRINT_TEXT))

        text = "(Conforme Lei Fed. 12.741/2012)"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))

        return json.dumps(report, default=lambda x: x.__dict__)

    def print_cancel(self, posid, data):
        return

    def _cancel_report(self, sat_fiscal_response, nCFe, total):
        return

    def _fiscal_report(self, order, tenders, sat_fiscal_response, nCFe, customer_name, model=None):
        # type: (eTree, List[Tender], eTree.Element, dict, str, str, Optional[eTree]) -> List[AposA8PrintLine]
        report = []

        for line in self.cliche_helper.get_cliche():
            report.append(AposA8PrintLine(AposA8Align.CENTER, line, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, "EXTRATO No. %s" % nCFe, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, "CUPOM FISCAL ELETRONICO - SAT", AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        self._append_coupon_body(
            customer_name=customer_name,
            model=model,
            order=order,
            report=report,
            tenders=tenders,
            printer_common=self.printer_common,
        )

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        # Campo de texto customizado exibido na área de informações adicionais
        if self.customer_text_field != "":
            for customer_text in self.customer_text_field.split('\\n'):
                report.append(AposA8PrintLine(AposA8Align.CENTER,
                                              customer_text.replace("\\n", ""),
                                              AposA8CommandType.PRINT_TEXT))
            report.append(
                AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        satk = insert_space(sat_fiscal_response.find("satkey").text.strip(), 4)
        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      str(satk),
                                      AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        cpf_cnpj = sat_fiscal_response.find("CustomerCPF").text
        cpf_text = "Consumidor: %s" % ("Nao informado" if not cpf_cnpj else cpf_cnpj)
        report.append(AposA8PrintLine(AposA8Align.CENTER, cpf_text, AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      "SAT No. %s" % sat_fiscal_response.find("satno").text.strip(),
                                      AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      sat_fiscal_response.find("datetime").text.strip(),
                                      AposA8CommandType.PRINT_TEXT))

        return report



import json

from mw_helper import ModelService
from sysactions import (
    action,
    get_model,
    StopAction,
    show_messagebox,
    get_podtype,
)
from typing import (
    Any,
    Union
)

from actions.util import check_sangria
from application.interactor import (
    CreateTableOrderInteractor,
    TableInteractor,
)
from custom_sysactions import block_action_if_pos_is_blocked


@action
@block_action_if_pos_is_blocked
def create_table_order(
    pos_id,             # type: str
    table_id,           # type: str
):
    # type: (...) -> str

    from msgbusboundary import (
        MsgBusOrderManager,
        MsgBusTableService as tableService,
    )
    from msgbusboundary.ui import MsgBusScreenChanger

    model = get_model(pos_id)
    check_sangria(model, pos_id)
    order_manager = MsgBusOrderManager()
    screen_changer = MsgBusScreenChanger()

    pod_type = get_podtype(model=model)
    model_service = ModelService(pos_id=pos_id)
    price_list = model_service.get_price_list_from_model(
        pod_type=pod_type,
        sale_type="EAT_IN",
    )

    table_service_period_dict = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="SERVICE_PERIOD_TABLE",
        table_id=table_id,
    )

    if table_service_period_dict:
        table_service_period_json = table_service_period_dict.get("SERVICE_PERIOD_TABLE", "{}")
        table_service_period = json.loads(table_service_period_json).get("servicePeriod")
        if table_service_period is not None:
            price_list = "{}.{}".format(table_service_period, price_list)

    if not tableService().is_a_tab(pos_id, table_id):
        table_id = _select_table_or_related_tab_by_dialog(
            pos_id=pos_id,
            table_id=table_id,
        )

    create_table_order_interactor = CreateTableOrderInteractor(
        table_service=tableService(),
        screen_changer=screen_changer,
        order_manager=order_manager,
    )

    status_order = create_table_order_interactor.execute(
        pos_id=pos_id,
        table_id=table_id,
        price_list=price_list,
    )

    return status_order


def _select_table_or_related_tab_by_dialog(
        pos_id,   # type: Any[Union[str, int]]
        table_id  # type: Any[Union[str, int]]
):
    # type: (...) -> Any[Union[str, int]]

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import (
        deselect_table,
        select_table_or_tab,
    )

    table_interactor = TableInteractor(pos_id, table_id, tableService())
    related_services, totaled_services = table_interactor.get_related_services(
        filter_totaled=True
    )

    if not related_services and totaled_services:
        show_messagebox(pos_id, "$HAS_NO_OPENED_TABS", "$INFORMATION")
        raise StopAction()
    selected_table_id = table_interactor.get_selected_related_services(related_services)
    if selected_table_id != table_id:
        deselect_table(pos_id)
        ret = select_table_or_tab(pos_id, selected_table_id)
        if ret:
            return selected_table_id
        else:
            raise StopAction()

    return selected_table_id
